@@ -2,6 +2,14 @@
 Workflows Client
 """
 from relevanceai.utils.base import _Base
+from enum import Enum
+from typing import Union
+
+
+class WorkflowStatus(Enum):
+    IN_PROGRESS: str = "inprogress"
+    COMPLETED: str = "complete"
+    FAILED: str = "failed"
 
 
 class WorkflowsClient(_Base):
@@ -53,3 +61,28 @@ class WorkflowsClient(_Base):
             method="POST",
             parameters={"metadata": metadata},
         )
+
+    def status(
+        self,
+        workflow_id: str,
+        metadata: dict,
+        workflow_name: str,
+        additional_information: str = "",
+        status: Union[WorkflowStatus, str] = WorkflowStatus.IN_PROGRESS,
+    ):
+        """
+        If status is complete, it triggers an email.
+        """
+        if isinstance(status, WorkflowStatus):
+            status = status.value
+        return self.make_http_request(
+            f"/workflows/{workflow_id}/status",
+            method="POST",
+            parameters={
+                "metadata": {},
+                "status": status,
+                "workflow_name": workflow_name,
+                "additional_information": additional_information,
+                "metadata": metadata,
+            },
+        )
