@@ -237,6 +237,7 @@ class Operations(Write):
             output_fields=output_fields,
             **kwargs,
         )
+
         return ops
 
     @track
@@ -628,9 +629,13 @@ class Operations(Write):
             if key in kwargs:
                 run_kwargs[key] = kwargs.pop(key)
 
+        if "credentials" not in kwargs:
+            kwargs["credentials"] = self.credentials
+
         cluster_ops = BatchClusterOps(
             model=model,
             alias=alias,
+            dataset_id=self.dataset_id,
             vector_fields=vector_fields,
             model_kwargs=model_kwargs,
             **kwargs,
@@ -741,6 +746,7 @@ class Operations(Write):
             refresh=refresh,
             **kwargs,
         )
+
         return ops
 
     def apply_transformers_pipeline(
@@ -1044,6 +1050,7 @@ class Operations(Write):
             refresh=refresh,
             **kwargs,
         )
+
         return ops
 
     def analyze_text(
