@@ -1,12 +1,21 @@
 """
 Base class for base.py to inherit.
 All functions related to running operations on datasets.
+
+The Pull Transform Push library is designed to be able to consistently
+data from Relevance AI Database, transform and then constantly push data 
+to the Relevance AI Database. This ensures that resources are utilised 
+to their limits.
+
 """
+import sys
+import time
 import psutil
 import threading
 import multiprocessing as mp
 import warnings
 
+from copy import deepcopy
 from datetime import datetime
 from typing import Any, Dict, List, Tuple, Type, Union, Optional, Callable
 from relevanceai.constants.constants import CONFIG
@@ -18,25 +27,37 @@ from tqdm.auto import tqdm
 
 from relevanceai.utils.helpers.helpers import getsizeof
 
+KILL_SIGNAL = "_KILL_QUEUE_SIGNAL_"
+
 
 class PullTransformPush:
 
+    pull_count: int
+    transform_count: int
+    push_count: int
+
     pull_bar: tqdm
     transform_bar: tqdm
     push_bar: tqdm
 
     pull_thread: threading.Thread
-    update_threads: List[threading.Thread]
+    transform_threads: List[threading.Thread]
     push_threads: List[threading.Thread]
 
+    pull_dataset: Dataset
+    push_dataset: Dataset
+    _has_kill_signal: bool = True
+
     def __init__(
         self,
-        dataset: Dataset,
-        func: Callable,
+        dataset: Optional[Dataset] = None,
+        pull_dataset: Optional[Dataset] = None,
+        func: Optional[Callable] = None,
+        push_dataset: Optional[Dataset] = None,
         func_args: Optional[Tuple[Any, ...]] = None,
         func_kwargs: Optional[Dict[str, Any]] = None,
         multithreaded_update: bool = False,
-        pull_chunksize: Optional[int] = 128,
+        pull_chunksize: Optional[int] = None,
         warmup_chunksize: Optional[int] = None,
         transform_chunksize: Optional[int] = 128,
         push_chunksize: Optional[int] = None,
@@ -46,7 +67,9 @@ class PullTransformPush:
         push_workers: int = 1,
         buffer_size: int = 0,
         show_progress_bar: bool = True,
-        timeout: Optional[int] = None,
+        show_pull_progress_bar: bool = True,
+        show_transform_progress_bar: bool = True,
+        show_push_progress_bar: bool = True,
         ingest_in_background: bool = False,
         run_in_background: bool = False,
         ram_ratio: float = 0.8,
@@ -54,6 +77,7 @@ class PullTransformPush:
         retry_count: int = 3,
         after_id: Optional[List[str]] = None,
         pull_limit: Optional[int] = None,
+        timeout: Optional[int] = None,
     ):
         """
         Buffer size:
@@ -62,40 +86,59 @@ class PullTransformPush:
         """
         super().__init__()
 
-        self.dataset = dataset
-        self.dataset_id = dataset.dataset_id
+        if dataset is None:
+            self.push_dataset = push_dataset  # type: ignore
+            self.pull_dataset = pull_dataset  # type: ignore
+
+        else:
+            self.push_dataset = dataset
+            self.pull_dataset = dataset
+
+        if dataset is None and pull_dataset is None and push_dataset is None:
+            raise ValueError(
+                "Please set `dataset=` or `push_dataset=` and `pull_dataset=`"
+            )
+
+        self.pull_dataset_id = self.pull_dataset.dataset_id
+        self.push_dataset_id = self.push_dataset.dataset_id
+
         self.config = CONFIG
 
         self.pull_limit = pull_limit
-        ndocs = self.dataset.get_number_of_documents(
-            dataset_id=self.dataset_id,
-            filters=filters,
-        )
+
+        self.failed_documents_count = 0
+
         if pull_limit is None:
+            ndocs = self.pull_dataset.get_number_of_documents(
+                dataset_id=self.pull_dataset_id,
+                filters=filters,
+            )
             self.ndocs = ndocs
         else:
             self.ndocs = pull_limit
 
         self.pull_chunksize = pull_chunksize
-        self.transform_chunksize = min(transform_chunksize, ndocs)
+        self.transform_chunksize = min(transform_chunksize, self.ndocs)
         self.warmup_chunksize = warmup_chunksize
         self.push_chunksize = push_chunksize
 
         self.update_all_at_once = update_all_at_once
         if update_all_at_once:
-            self.transform_chunksize = ndocs
+            self.transform_chunksize = self.ndocs
 
-        tqdm.write(f"Transform Chunksize: {self.transform_chunksize:,}")
+        if func is None:
+            tqdm.write(f"Transform Chunksize: {self.transform_chunksize:,}")
 
-        self.timeout = 30 if timeout is None else timeout
         self.ingest_in_background = ingest_in_background
 
         self.filters = [] if filters is None else filters
         self.select_fields = [] if select_fields is None else select_fields
 
         self.general_lock = threading.Lock()
+
         self.transform_batch_lock = threading.Lock()
         self.push_batch_lock = threading.Lock()
+
         self.func_lock: Union[threading.Lock, None]
 
         if not multithreaded_update:
@@ -111,7 +154,7 @@ class PullTransformPush:
         self.func_kwargs = {} if func_kwargs is None else func_kwargs
 
         if update_all_at_once:
-            self.single_queue_size = ndocs
+            self.single_queue_size = self.ndocs
 
         else:
             if buffer_size == 0:
@@ -129,17 +172,49 @@ class PullTransformPush:
 
         tqdm.write(f"Max number of documents in queue: {self.single_queue_size:,}")
 
-        self.tq: mp.Queue = mp.Queue(maxsize=self.single_queue_size)
-        self.pq: mp.Queue = mp.Queue(maxsize=self.single_queue_size)
+        # mp queues are thread-safe while being able to be used across processes
+        self.tq: mp.JoinableQueue = mp.JoinableQueue(maxsize=self.single_queue_size)
+        self.pq: mp.JoinableQueue = mp.JoinableQueue(maxsize=self.single_queue_size)
         self.func = func
 
-        self.tqdm_kwargs = dict(leave=True, disable=(not show_progress_bar))
+        self.pull_tqdm_kwargs = dict(
+            leave=True,
+            disable=(not (show_pull_progress_bar and show_progress_bar)),
+        )
+        if not self.pull_tqdm_kwargs["disable"]:
+            self.pull_tqdm_kwargs["position"] = 0  # type: ignore
+
+        self.transform_tqdm_kwargs = dict(
+            leave=True,
+            disable=(not (show_transform_progress_bar and show_progress_bar)),
+        )
+        if not self.transform_tqdm_kwargs["disable"]:
+            self.transform_tqdm_kwargs["position"] = (
+                self.pull_tqdm_kwargs.get("position", 0) + 1  # type: ignore
+            )
+
+        self.push_tqdm_kwargs = dict(
+            leave=True,
+            disable=(not (show_push_progress_bar and show_progress_bar)),
+        )
+        if not self.push_tqdm_kwargs["disable"]:
+            self.push_tqdm_kwargs["position"] = (
+                self.pull_tqdm_kwargs.get("position", 0)  # type: ignore
+                + self.transform_tqdm_kwargs.get("position", 0)
+                + 1
+            )
+
         self.run_in_background = run_in_background
 
         self.failed_frontier: Dict[str, int] = {}
         self.retry_count = retry_count
         self.after_id = after_id
 
+        # time limit is 1hr 50mins if not set, this is to leave 10mins at the end
+        # of sagemaker job to send workflow status email
+        self.timeout = timeout
+        self.timeout_event = threading.Event()
+
     def _get_average_document_size(self, sample_documents: List[Dict[str, Any]]):
         """
         Get average size of a document in memory.
@@ -172,10 +247,34 @@ class PullTransformPush:
         documents: List[Dict[str, Any]] = [{"placeholder": "placeholder"}]
         after_id: Union[List[str], None] = self.after_id
 
-        while documents:
-            res = self.dataset.datasets.documents.get_where(
-                dataset_id=self.dataset_id,
-                page_size=20 if self.pull_chunksize is None else self.pull_chunksize,
+        while not self.timeout_event.is_set():
+            current_count = self.pull_bar.n
+
+            if self.pull_chunksize is None:
+                # this is the first batch
+                pull_chunksize = 20
+            else:
+                # optimized pull batch size (every other pull)
+                pull_chunksize = self.pull_chunksize
+
+            if self.pull_limit is None:
+                # Very large number if no limits
+                pull_limit = sys.maxsize
+            else:
+                # if there is a limit, get the ndocs left
+                pull_limit = self.pull_limit - current_count
+
+            # Consider all scenarios
+            # let 3333 = pull_limit, current_count = 3000
+            #         = min(20,            sys.maxsize)             = 20
+            #         = min(20,            (3333 - 3000 = 333))     = 20
+            #         = min(~512,          sys.maxsize)             = 512
+            #         = min(~512,          (3333 - 3000 = 333))     = 333
+            page_size = min(pull_chunksize, pull_limit)
+
+            res = self.pull_dataset.datasets.documents.get_where(
+                dataset_id=self.pull_dataset_id,
+                page_size=page_size,
                 filters=self.filters,
                 select_fields=self.select_fields,
                 sort=[],
@@ -183,6 +282,12 @@ class PullTransformPush:
             )
 
             documents = res["documents"]
+            if not documents:
+                with self.general_lock:
+                    self.ndocs = self.pull_count
+                print("Killing transform queue.")
+                self.tq.put(KILL_SIGNAL)
+                break
             after_id = res["after_id"]
 
             if self.pull_chunksize is None:
@@ -195,33 +300,29 @@ class PullTransformPush:
 
             with self.general_lock:
                 self.pull_bar.update(len(documents))
+                self.pull_count += len(documents)
 
-            if self.pull_limit is not None:
-                if self.pull_bar.n > self.pull_limit:
-                    break
-
-    def _get_update_batch(self) -> List[Dict[str, Any]]:
+    def _get_transform_batch(self) -> List[Dict[str, Any]]:
         """
         Collects a batches of of size `transform_chunksize` from the transform queue
         """
         batch: List[Dict[str, Any]] = []
 
         queue = self.tq
-        timeout = 5
+        timeout = 5 if not self.update_all_at_once else None
 
-        if self.transform_bar.n == 0 and self.warmup_chunksize is not None:
+        if self.transform_count == 0 and self.warmup_chunksize is not None:
             chunksize = self.warmup_chunksize
             tqdm.write("Processing Warmup Batch")
         else:
             chunksize = self.transform_chunksize
 
-        while self.update_all_at_once or len(batch) < chunksize:
+        while len(batch) < chunksize:
             try:
                 document = queue.get(timeout=timeout)
                 batch.append(document)
             except:
-                if len(batch) > 0:
-                    break
+                break
 
         return batch
 
@@ -236,7 +337,14 @@ class PullTransformPush:
 
         # Calculate optimal batch size
         if self.push_chunksize is None:
-            sample_documents = [queue.get(timeout=timeout) for _ in range(10)]
+            sample_documents = []
+            for _ in range(10):
+                try:
+                    sample_document = queue.get(timeout=timeout)
+                except:
+                    break
+                sample_documents.append(sample_document)
+
             self.push_chunksize = self._get_optimal_chunksize(sample_documents, "push")
             batch = sample_documents
 
@@ -244,9 +352,9 @@ class PullTransformPush:
         while len(batch) < chunksize:
             try:
                 document = queue.get(timeout=timeout)
+                batch.append(document)
             except:
                 break
-            batch.append(document)
 
         return batch
 
@@ -257,25 +365,54 @@ class PullTransformPush:
         ^ necessary to avoid reinserting stuff that is already in the cloud.
         Then, repeatedly put each document from the processed batch in the push queue
         """
-        while self.transform_bar.n < self.ndocs:
+        # Check for early termination (such as no documents)
+        HAS_KILL_SIGNAL: bool = False
+        print("Begin transform.")
+        while self.transform_count <= self.ndocs and not self.timeout_event.is_set():
+            print("Inside transform loop.")
             with self.transform_batch_lock:
-                batch = self._get_update_batch()
+                batch = self._get_transform_batch()
+                if len(batch) > 0:
+                    if batch[-1] == KILL_SIGNAL:
+                        HAS_KILL_SIGNAL = True
+                        print("Killing transform queue.")
+                        self.tq.task_done()
+                        batch = batch[:-1]
 
-            old_keys = [set(document.keys()) for document in batch]
+            if self.func is not None:
+                old_batch = deepcopy(batch)
 
-            if self.func_lock is not None:
-                with self.func_lock:
-                    new_batch = self.func(batch, *self.func_args, **self.func_kwargs)
-            else:
-                new_batch = self.func(batch, **self.func_kwargs)
+                if self.func_lock is not None:
+                    with self.func_lock:
+                        try:
+                            new_batch = self.func(
+                                batch, *self.func_args, **self.func_kwargs
+                            )
+                        except Exception as e:
+                            print(e)
+                            new_batch = batch
+                else:
+                    try:
+                        new_batch = self.func(batch, **self.func_kwargs)
+                    except Exception as e:
+                        print(e)
+                        new_batch = batch
 
-            batch = PullTransformPush._postprocess(new_batch, old_keys)
+                batch = PullTransformPush._postprocess(new_batch, old_batch)
 
             for document in batch:
                 self.pq.put(document)
 
+            # Send kill signal to push queue
+            if HAS_KILL_SIGNAL:
+                print("Killing Push queue")
+                self.pq.put(KILL_SIGNAL)
+
             with self.general_lock:
                 self.transform_bar.update(len(batch))
+                self.transform_count += len(batch)
+            if HAS_KILL_SIGNAL:
+                sys.exit()
 
     def _handle_failed_documents(
         self,
@@ -291,8 +428,8 @@ class PullTransformPush:
 
         if failed_documents:
             with self.general_lock:
-                self.ndocs += len(failed_documents)
-                desc = f"push - failed_documents = {self.ndocs - self.ndocs}"
+                self.failed_documents_count += len(failed_documents)
+                desc = f"push - failed_documents = {self.failed_documents_count}"
                 self.push_bar.set_description(desc)
 
             # ...find these failed documents within the batch...
@@ -317,20 +454,33 @@ class PullTransformPush:
     @staticmethod
     def _postprocess(
         new_batch: List[Dict[str, Any]],
-        old_keys: List[str],
+        old_batch: List[Dict[str, Any]],
     ) -> List[Dict[str, Any]]:
         """
         Removes fields from `new_batch` that are present in the `old_keys` list.
         Necessary to avoid bloating the upload payload with unnecesary information.
         """
-        batch = [
-            {
-                key: value
-                for key, value in new_batch[idx].items()
-                if key not in old_keys[idx] or key == "_id"
-            }
-            for idx in range(len(new_batch))
-        ]
+        batch = []
+        for old_document, new_document in zip(old_batch, new_batch):
+            document: Dict[str, Any] = {}
+            new_fields = Dataset.list_doc_fields(doc=new_document)
+            old_fields = Dataset.list_doc_fields(doc=old_document)
+            for field in new_fields:
+                old_value = Dataset.get_field(
+                    field=field,
+                    doc=old_document,
+                    missing_treatment="return_none",
+                )
+                new_value = Dataset.get_field(
+                    field=field,
+                    doc=new_document,
+                    missing_treatment="return_none",
+                )
+                value_diff = old_value != new_value
+                if field not in old_fields or value_diff or field == "_id":
+                    Dataset.set_field(field=field, doc=document, value=new_value)
+            batch.append(document)
+
         return batch
 
     @staticmethod
@@ -347,16 +497,24 @@ class PullTransformPush:
         """
         Iteratively gather a batch of processed documents and push these to cloud
         """
-        while self.push_bar.n < self.ndocs:
+        # Ensure kill signal is only sent after the transformation
+        # is done and the push is added
+        HAS_KILL_SIGNAl = False
+        while self.push_count < self.ndocs and not self.timeout_event.is_set():
             with self.push_batch_lock:
                 batch = self._get_push_batch()
+                if len(batch) > 0:
+                    if batch[-1] == KILL_SIGNAL:
+                        HAS_KILL_SIGNAl = True
+                        batch = batch[:-1]
+                        self.pq.task_done()
 
-            batch = self.dataset.json_encoder(batch)
+            batch = self.pull_dataset.json_encoder(batch)
             update = PullTransformPush._get_updates(batch)
 
             if update:
-                res = self.dataset.datasets.documents.bulk_update(
-                    self.dataset_id,
+                res = self.push_dataset.datasets.documents.bulk_update(
+                    self.push_dataset_id,
                     batch,
                     return_documents=True,
                     ingest_in_background=self.ingest_in_background,
@@ -372,79 +530,163 @@ class PullTransformPush:
 
             with self.general_lock:
                 self.push_bar.update(len(batch) - len(failed_documents))
+                self.push_count += len(batch) - len(failed_documents)
+
+            # tell push queue to also finish
+            if HAS_KILL_SIGNAl:
+                self.pq.task_done()
+                # Close threads
+                sys.exit()
 
     def _init_progress_bars(self) -> None:
         """
         Initialise the progress bars for dispay progress on pulling updating and pushing.
         """
+        self.pull_count = 0
         self.pull_bar = tqdm(
             desc="pull",
-            position=0,
             total=self.ndocs,
-            **self.tqdm_kwargs,
+            **self.pull_tqdm_kwargs,
         )
+        self.transform_count = 0
         self.transform_bar = tqdm(
             range(self.ndocs),
             desc="transform",
-            position=1,
-            **self.tqdm_kwargs,
+            **self.transform_tqdm_kwargs,
         )
+        self.push_count = 0
         self.push_bar = tqdm(
             range(self.ndocs),
             desc="push",
-            position=2,
-            **self.tqdm_kwargs,
+            **self.push_tqdm_kwargs,
         )
 
     def _init_worker_threads(self) -> None:
         """
         Initialise the worker threads for each process
         """
-        self.pull_thread = threading.Thread(target=self._pull)
-        self.update_threads = [
-            threading.Thread(target=self._transform)
+        daemon = True if self.timeout is not None else False
+        self.pull_thread = threading.Thread(target=self._pull, daemon=daemon)
+        self.transform_threads = [
+            threading.Thread(target=self._transform, daemon=daemon)
             for _ in range(self.transform_workers)
         ]
         self.push_threads = [
-            threading.Thread(target=self._push) for _ in range(self.push_workers)
+            threading.Thread(target=self._push, daemon=daemon)
+            for _ in range(self.push_workers)
         ]
 
-    def _run_worker_threads(self):
+    def _start_worker_threads(self):
         """
-        Start the worker threads and then join them in reversed order.
+        Start the worker threads
         """
+
+        # Start fetching data from the server
         self.pull_thread.start()
-        while True:
-            if not self.tq.empty():
-                for thread in self.update_threads:
-                    thread.start()
-                break
-        while True:
-            if not self.pq.empty():
-                for thread in self.push_threads:
-                    thread.start()
-                break
+        # Once there is data in the queue, then we start the
+        # transform threads
+        while self.tq.empty():
+            time.sleep(1)
+        for thread in self.transform_threads:
+            thread.start()
+        while self.pq.empty():
+            time.sleep(1)
+        for thread in self.push_threads:
+            thread.start()
 
+    def _join_worker_threads(self, timeout: Optional[int] = None):
+        """
+        ...and then join them in reversed order.
+        """
+
+        # Try to join if not running in background
         if not self.run_in_background:
+            self.pull_thread.join(timeout=timeout)
+            for thread in self.transform_threads:
+                thread.join(timeout=timeout)
             for thread in self.push_threads:
-                thread.join()
-            for thread in self.update_threads:
-                thread.join()
-            self.pull_thread.join()
+                thread.join(timeout=timeout)
 
-    def run(self) -> List[str]:
+    def _threads_are_alive(self) -> True:
+        """
+        Poll all active trheads to check if they are alive
+        """
+        if self.pull_thread.is_alive():
+            return True
+        for thread in self.transform_threads:
+            if thread.is_alive():
+                return True
+        for thread in self.push_threads:
+            if thread.is_alive():
+                return True
+        return False
+
+    def _run_timer(self):
+        start_time = time.time()
+        while True:
+            current_time = time.time()
+
+            # check if time limit was exceeded
+            if (current_time - start_time) >= self.timeout:
+                tqdm.write("Time Limit Exceeded")
+                with self.general_lock:
+                    self.timeout_event.set()
+                tqdm.write("Exiting Operation...")
+                break
+
+            # or if all the threads have finished
+            if not self._threads_are_alive():
+                break
+
+            # poll these checks every 1 sec
+            time.sleep(1.0)
+
+    def _flush_queues(self, timeout: float = 1e-2):
+        """
+        Gets all items in both queues to avoid
+        BrokenPipeError when calling queue.close()
+        """
+        while True:
+            try:
+                self.tq.get(timeout=timeout)
+            except:
+                break
+
+        while True:
+            try:
+                self.pq.get(timeout=timeout)
+            except:
+                break
+
+        self.tq.close()
+        self.pq.close()
+
+    def run(self) -> Dict[str, Any]:
         """
         (Main Method)
         Do the pulling, the updating, and of course, the pushing.
 
         return the _ids of any failed documents
         """
+
         if self.ndocs > 0:
             self._init_progress_bars()
             self._init_worker_threads()
-            self._run_worker_threads()
+            self._start_worker_threads()
 
-        return list(self.failed_frontier.keys())
+            if self.timeout is None:
+                self._join_worker_threads()
+
+            else:
+                self._run_timer()  # Starts the timer
+                self._flush_queues()
+                # no need to join threads as they are daemons
+                # if there is a timeout
+
+        return {
+            "timed_out": self.timeout_event.is_set(),
+            "failed_documents": list(self.failed_frontier.keys()),
+        }
 
 
 def arguments(cls: Type[PullTransformPush]):
@@ -539,7 +781,7 @@ class OperationRun(TransformBase):
             dataset=dataset,
             operation=self,
         ) as dataset:
-            self.batch_transform_upsert(
+            res = self.batch_transform_upsert(
                 dataset=dataset,
                 select_fields=select_fields,
                 filters=filters,
@@ -548,7 +790,7 @@ class OperationRun(TransformBase):
                 **kwargs,
             )
 
-        return
+        return res
 
     def batch_transform_upsert(
         self,
@@ -560,7 +802,7 @@ class OperationRun(TransformBase):
         chunksize: int = None,
         transform_workers: int = 2,
         push_workers: int = 2,
-        timeout: int = 30,
+        timeout: int = None,
         buffer_size: int = 0,
         show_progress_bar: bool = True,
         warmup_chunksize: int = None,
