@@ -1,3 +1,4 @@
+import time
 import pytest
 
 from typing import Dict, List
@@ -10,6 +11,8 @@ from tests.globals.constants import generate_dataset_id
 @pytest.fixture(scope="function")
 def vector_dataset_id(test_client: Client, vector_documents: List[Dict]):
     test_dataset_id = generate_dataset_id()
-    test_client._insert_documents(test_dataset_id, vector_documents)
+    test_dataset = test_client.Dataset(test_dataset_id)
+    test_dataset.insert_documents(vector_documents)
+    time.sleep(1)
     yield test_dataset_id
     test_client.datasets.delete(test_dataset_id)
